/*********************************************************************************************************\
 |                                           Juego del buscaminas                                          |
 +---------------------------------------------------------------------------------------------------------+
 |  Ejemplo de implementación del juego buscaminas, utiliza gráficos muy sencillos                         |
 |                                                                                                         |
 |  Se crean 2 clases Tablero y Celda                                                                      |
 |                                                                                                         |
 |  La clase Tablero representa el tablero que se mostrará en la pantalla, está fomado por un conjunto de  |
 |  celdas.                                                                                                |
 |                                                                                                         |
 |  La clase Celda representa cada una de las celdas del tablero, debajo de ella puede haber una bomba     |
 |                                                                                                         |
 \*********************************************************************************************************/

// Constantes
static final int ancho = 30; // Ancho fijo de las celdas
static final int BOMBA = -1;
static final int VACIO = 0;

// Maquina de estados
static final int INSTRUCCIONES = 0;
static final int JUGANDO = 1;
static final int GANADOR = 2;
static final int GAMEOVER = 3;

int estado = INSTRUCCIONES;

// Variables del tablero de juego
int num_celdas = 10;
Tablero tablero;

// Fuente de los mensajes
PFont font16, font12; 

/** Inicialización del tamaño del lienzo por parámetros */
void settings() {
  size(ancho*num_celdas, ancho*num_celdas+100);
}

/** Inicialización las variables y el lienzo */
void setup() {
  background(0);
  // Creamos la fuente para los textos
  font16 = createFont("Consolas", 16, true);
  font12 = createFont("Consolas", 12, true);

  textAlign(CENTER);

  // Creamos el tablero
  tablero = new Tablero(num_celdas, 5);
}

/**
 Pintado del lienzo dependiendo del estado
 */
void draw() {
  background(0);
  // Pintamos el texto con los créditos de autoría
  fill(255, 255, 0);
  textFont(font12, 12);
  text("CC-BY Pedro Puertas Estivill", width/2, height-20);  
  // Dependiendo de la máquina de estados pintamos las instrucciones, el tablero o 
  // el indicador de que ha ganado o perdido
  switch (estado) {
  case INSTRUCCIONES:
    fill(255);
    textFont(font16, 16);
    text("Buscaminas", width/2, 20);
    textFont(font12, 12);
    text("Debes encontrar las "+tablero.num_bombas+" bombas que ", width/2, 40);
    text("hay escondidas en el tablero", width/2, 52);
    text("Pulsa una tecla para empezar", width/2, 72);
    return;
  case GANADOR: 
    fill(0, 155, 0);
    textFont(font16, 16);
    text("BIEN!!! HAS GANADO", width/2, height-70);
    fill(255);
    textFont(font12, 12);
    text("Pulsa una tecla para otra partida", width/2, height-50);
    break;
  case GAMEOVER: 
    fill(255, 0, 0);
    textFont(font16, 16);
    text("GAMEOVER", width/2, height-70);
    fill(255);
    textFont(font12, 12);
    text("Pulsa una tecla para otra partida", width/2, height-50);
    break;
  }
  // Siempre pintamos el tablero excepto en las instrucciones que vuelve directamente
  tablero.draw();
}

/**
 Control del teclado
 */
void keyPressed() {
  //println(key);
  //println(keyCode);
  if (key == CODED) return; 
  // Dependiendo del estado actual hacemos una acción u otra 
  switch (estado) {
  case INSTRUCCIONES:
    estado = JUGANDO; // Pasamos a jugando si está mostrando las instrucciones
    return;
  case GANADOR:  
  case GAMEOVER:
    // Reiniciamos el tablero para iniciar una nueva partida
    int nb = 2 + (int)random(8);
    num_celdas = 5 + nb;
    if (num_celdas < 10) num_celdas = 10;
    tablero = new Tablero(num_celdas, nb*3);
    // Cambiamos el tamaño de la pantalla que es aleatorio.
    surface.setSize(ancho*num_celdas, ancho*num_celdas+100);
    estado = JUGANDO;
    return;
  }
}

void mousePressed() {
  if (estado != JUGANDO) return;
  if (mouseButton == RIGHT) {
    tablero.marca(mouseX, mouseY);
    return;
  }  
  if (tablero.destapa(mouseX, mouseY) == BOMBA) {
    estado = GAMEOVER;
    tablero.mostrar();
  }
  if (tablero.haGanado()) estado = GANADOR;
}  


class Tablero {
  private int filas;
  private int columnas;
  private int num_bombas;
  private int no_visibles;

  Celda tablero[][];

  public Tablero(int r, int b) {
    this.filas = r;
    this.columnas = r;
    this.num_bombas = b;
    iniciar_tablero();
  }

  void iniciar_tablero() {
    if (tablero == null) {
      println("Iniciando el tablero");
      no_visibles = columnas*filas;
      tablero = new Celda[columnas][filas];

      for (int j=0; j < filas; j++) {
        for (int i=0; i < columnas; i++) {
          tablero[i][j] = new Celda(i, j);
        }
      }
    }
    int n = num_bombas;
    while (n > 0) {
      int i = (int)random(columnas);
      int j = (int)random(filas);
      if (tablero[i][j].estado != BOMBA) {
        tablero[i][j].estado = BOMBA;
        //tablero[i][j].visible = true;
        n--;
      }
    }

    for (int j=0; j < filas; j++) {
      for (int i=0; i < columnas; i++) {
        tablero[i][j].estado = calcula_estado(i, j);
      }
    }
  }

  int estado_celda(int i, int j) {
    return tablero[i][j].estado;
  }

  void draw() {
    textFont(font16, 16);
    for (int j=0; j < filas; j++) {
      for (int i=0; i < columnas; i++) {
        tablero[i][j].draw();
      }
    }
  }

  void mostrar() {
    for (int j=0; j < filas; j++) {
      for (int i=0; i < columnas; i++) {
        tablero[i][j].visible = true;
      }
    }
  }

  void marca(int x, int y) {
    int i = x/ancho;
    int j = y/ancho;
    if (i >= 0 && i < columnas && j >= 0 && j < filas) {
      if (!tablero[i][j].visible) tablero[i][j].marcada = !tablero[i][j].marcada;
    }
  }

  int destapa(int x, int y) {
    // Calculamos a partir de las coordenadas de la pantalla que celda hay que destapar
    int i = x/ancho;
    int j = y/ancho;
    
    // Nos aseguramos que la celda calculada esté dentrol del rango del tablero
    if (i >= 0 && i < columnas && j >= 0 && j < filas) {
      // Si la celda está marcada volvemos
      if (tablero[i][j].marcada) return -2;
      // Si hay una bomba o ya está destapada retornamos el estado de la celda
      if (tablero[i][j].estado == BOMBA || tablero[i][j].visible) return tablero[i][j].estado;

      // Si la celda no está vacia y no es una bomba, destapamos la celda, descontamos las celdas
      // pendientes para controlar si el usuario ha ganado y retornamos su valor
      if (tablero[i][j].estado > VACIO) {
        tablero[i][j].visible = true;
        no_visibles--;
        return tablero[i][j].estado;
      }  
      // Si la celda está vacía, destaparemos la celda clicada y todas las celdas adyacentes
      if (tablero[i][j].estado == VACIO) {
        tablero[i][j].visible = true;
        no_visibles--;
        destapa(x-ancho, y-ancho); // Destapar esquina superior izquierda
        destapa(x, y-ancho);       // Destapar celda superior
        destapa(x+ancho, y-ancho); // Destapar esquina superior derecha
        destapa(x-ancho, y);       // Destapar celda izquierda
        destapa(x+ancho, y);       // Destapar celda derecha
        destapa(x-ancho, y+ancho); // Destapar celda inferior izquierda
        destapa(x, y+ancho);       // Destapar celda inferior
        destapa(x+ancho, y+ancho); // Destapar celda inferior derecha
      }
    }
    return VACIO;
  }

  boolean haGanado() {
    return no_visibles <= this.num_bombas;
  }


  private int calcula_estado(int i, int j) {
    if (tablero[i][j].estado == BOMBA) return BOMBA;

    int total_bombas = 0;
    for (int n=-1; n <= 1; n++) {
      for (int m=-1; m <= 1; m++) {
        if (i+m >= 0 && i+m < columnas && j+n >= 0 && j+n < filas) {
          if (tablero[i+m][j+n].estado == BOMBA) total_bombas++;
        }
      }
    }
    return total_bombas;
  }
}


class Celda {
  public int estado = VACIO;
  private int i, j;
  public boolean visible = false;  
  public boolean marcada = false; 

  public Celda(int i, int j) {
    this.i = i;
    this.j = j;
  }

  public void draw() {
    // Si visible fijamos el color de fondo a gris claro, en caso contrario será gris oscuro
    if (visible) fill(180, 180, 180); 
    else fill(60, 60, 60);
    // Lineas en negro
    stroke(0);
    // Pintamos el recuadro 
    rect(i*ancho, j*ancho, ancho, ancho); 
    
    if (visible) {
      // Analizamos el estado cuando la celda está visible
      switch(estado) {
      case BOMBA:
        // Si es una bomba la pintamos con un circulo rojo en medio
        fill(255, 0, 0);
        noStroke();
        ellipse((i*ancho)+ancho/2, (j*ancho)+ancho/2, ancho/2, ancho/2);
        return;
      case VACIO: 
        // Si está vacia 
        return;
      case 1:
        // Si hay una de las celdas adyacentes que es una bomba pintamos el texto
        // en verde
        fill(0, 155, 0);
        break;
      case 2:
        // Si hay dos celdas adyacentes que son bombas pintamos el textos en azul
        fill(0, 0, 255);
        break;
      default:  
        // Si hay más de dos celdas que son bombas el texto se pinta en rojo
        fill(255, 0, 0);
      } 
      // Pintamos el texto centrado en la celda
      text(estado, i*ancho+(ancho/2), j*ancho+(ancho/2)+8);
    } else {
      // Si el usuario ha marcado la celda, la pintamos como un circulo de color azul
      if (marcada) {
        fill(0, 0, 255);
        noStroke();
        ellipse((i*ancho)+ancho/2, (j*ancho)+ancho/2, ancho/2, ancho/2);
      }
    }  
  }
}