
/***********************************************************************************\
 .                                                                                  .
 .                          JUEGO DE TIRO AL TANQUE                                 .
 .                                                                                  .
 .----------------------------------------------------------------------------------.
 .                                                                                  .
 . En este caso se implementa un sencillo juego de disparar un proyectil, el cual   .
 . va cambiando de tamaño y angulo mientras viaja por el cielo.                     .
 .                                                                                  .
 . Aspectos relevantes que se pueden ver son:                                       .
 .   - Animación por rotación del cañon                                             .
 .   - Movimiento del proyectil mediante formulas físicas de tiro parabólico        .
 .   - Animación de personajes (sprites)                                            .
 .                                                                                  .
 .    Todas las imagenes y el sonido son de dominio público, llicència CC-0         .
 .                                                                                  .
 -----------------------------------------------------------------------------------.
 .                                                                                  .
 .                    CC-BY -  P E D R O   P U E R T A S                            .
 .                                                                                  .
 .                     MOSAIC 2017 - VIDEOJUEGOS CON PROCESSING                     .
 .                                                                                  .
 \**********************************************************************************/



// Importar la libreria de controles y de sonido
import controlP5.*; 
import ddf.minim.*;

// Declaración de constantes de la máquina de estados
static final int REPOSO = 0;
static final int DISPARANDO = 1;
static final int INSTRUCCIONES = 2;
static final int DESTRUIDO = 3;
static final int ERROR = 4;

// Tiempo de espera para que desaparezca la pantalla de avisos
static final int TIEMPO_ESPERA = 5;
static final float gr = 9.8; // Velocidad de la gravedad

// Objeto libreria P5 para controles en pantalla
ControlP5 p5;
// Objeto de sonido para la explosión
Minim minim;
AudioPlayer sboom;

Sprite rana;     // Sprite de la rana (animado)
Sprite tanque;   // Sprite del tanque   
Sprite nube;     // Sprite de las nubes

// Variables per el bala
float x;             // Posició horizontal del bala 
float y;             // Posició vertical del bala
float v_ini=20;      // Velocidad inicial 
float angulo=15;     // Angulo inicial
float t=0;           // Posició en el temps

// Variables de control del tanque
float vtank = 3.5;     // Velocidad a la que es mou el tanque

float contador = 5;         // Contador de tiempo
int intentos = 5;        // Intentos para destruir el tanque
int estado = INSTRUCCIONES; // Máquina de estados

// Variables de las diferentes imágenes
PImage bala = null;   // Bala
PImage base = null;   // Base del cañón
PImage fusil = null;  // Lanzadera del cañón

// Fondo
PImage desierto = null;
// Explosión
PImage boom = null;

void setup() {
  // Iniciamos las medidas de la pantalla
  size(1024, 600);
  // Iniciamos la libreria P5
  p5 = new ControlP5(this);
  // Cargamos el sonido del booooommm
  minim = new Minim(this);
  sboom = minim.loadFile("boom.mp3");

  // Cada 41.6ms se llama a draw() 24 frames por segundo 
  frameRate(24); 

  // Cargamos las imágenes
  bala = loadImage ("bala.png");
  base = loadImage ("base.png");
  fusil = loadImage ("fusil.png");

  desierto = loadImage("desierto.png");
  boom = loadImage("boom.png");

  // Fijamos los colores para todos los controles P5
  p5.setColorActive(color(255, 80, 80));
  p5.setColorForeground(color(255, 80, 80));
  p5.setColorBackground(color(80, 60, 60, 100));

  controlP5.Button boton_v_mas = p5.addButton("boton_v_mas"); // El control se llamará boton_v_mas
  boton_v_mas.setPosition(125, 14); // Posición en la pantalla
  boton_v_mas.setSize(10, 10);  // Medida del botón
  boton_v_mas.setCaptionLabel("+"); // Texto del botón

  controlP5.Button boton_v_menos = p5.addButton("boton_v_menos"); // El control se llamará boton_v_menos
  boton_v_menos.setPosition(85, 14); // Posición en la pantalla
  boton_v_menos.setSize(10, 10);  // Medida del botón
  boton_v_menos.setCaptionLabel("-"); // Texto del botón

  controlP5.Button boton_a_mas = p5.addButton("boton_a_mas"); // El control se llamará boton_a_mas
  boton_a_mas.setPosition(125, 34); // Posición en la pantalla
  boton_a_mas.setSize(10, 10);  // Medida del botón
  boton_a_mas.setCaptionLabel("+"); // Texto del botón

  controlP5.Button boton_a_menos = p5.addButton("boton_a_menos"); // El control se llamará boton_a_menos
  boton_a_menos.setPosition(85, 34); // Posición en la pantalla
  boton_a_menos.setSize(10, 10);  // Medida del botón
  boton_a_menos.setCaptionLabel("-"); // Texto del botón

  controlP5.Button boto_dispara = p5.addButton("boton_dispara"); // El control se llamará boto_dispara
  boto_dispara.setPosition(15, 80); // Posición en la pantalla
  boto_dispara.setSize(120, 20);  // Medida del botón
  boto_dispara.setCaptionLabel("DISPARA"); // Texto del botón

  // La rana es una imagen de 8 frames (4x2)
  rana = new Sprite(width/2, height-100, 4, 2, "rana.png");
  rana.voltea(); // Como inicialmente mira a la izquierda la volteamos para que mire hacia la derecha
  //rana.velocidad(0.8);
  rana.medida(32, 64); // Ajustamos el tamaño de la rana
  rana.limites(-140, width+140); // Fijamos los límites de movimiento horizontal  

  // La nube no tiene animación y por tanto sólo tiene un frame
  nube = new Sprite(20, 70, 1, 1, "nube.png");
  nube.medida(250, 100);
  nube.limites(-280, width+280); // Fijamos los límites mínimo y máximo del movimiento horizontal  

  // El tanque tiene 2 frames
  tanque = new Sprite((int)(width-178-random(300)), height-80, 2, 1, "tanque.png");
  tanque.limites (width/2+50, width+100);
}

/**
 Función principal de pintado
 */
void draw() {
  // Pintamos el fondo con la imagen del desierto
  image(desierto, 0, 0, width, height);
  nube.draw();

  // Control del contador de inicio del juego
  if (contador > 0) contador -= (1/frameRate); // Descontamos segundos utilizando el frameRate

  // Si llegamos a cero y no estamos ya en la pantalla de instruccions, error o acierto. Podemos empezar a disparar
  if (contador <= 1 && estado >= INSTRUCCIONES) {
    // Iniciamos los contadores de posición y de intentos
    contador = 0;
    intentos = 5;
    x = 0;
    estado = REPOSO;
    v_ini=20;  // Velocidad inicial 
    angulo=15; // Angulo inicial
  }
  // Pintamos la pantalla dependiendo del estado de la máquina de estados
  textAlign(CENTER, CENTER);
  textSize(16);

  switch (estado) {
  case INSTRUCCIONES : 
    fill(0, 0, 0, 208);  //color negro semi-transparente
    rect(width/2-300, height/2-120, 600, 240, 10);  
    fill(255, 255, 0);
    text("Has de destruir el tanque cambiando el ángulo y la velocidad\r\n a la que lanza el cañón\r\n\r\n"+
      "Flechas arriba / abajo para variar el ángulo\r\n"+
      "Flechas izquierda / derecha para variar la velocidad\r\n"+
      "ESPACIO para disparar\r\n\r\n"+
      "También puedes hacer servir los controles del cuadro de mando", 
      width/2, height/2);

    textSize(80);
    fill(55, 55, 0, 208);
    text((int)contador, width-50, 50);
    break;
  case DESTRUIDO :
    textAlign(CENTER, CENTER);
    textSize(20);
    fill(0, 0, 0, 208);  //color negro semi-transparent
    rect(width/2-250, height/2-75, 500, 150, 10);  
    fill(255, 255, 0);
    text("M U Y   B I E N !!!!\r\n\r\nHas conseguido destruir el tanque.", width/2, height/2);
    image(boom, x-boom.width/2, height-boom.height);
    break;
  case ERROR:
    textAlign(CENTER, CENTER);
    textSize(20);
    fill(55, 0, 0, 208);  //color negro semi-transparente
    rect(width/2-250, height/2-75, 500, 150, 10);  
    fill(255, 255, 0);
    text("OHHHHHHHHHH !!!!\r\n\r\nNO has conseguido acertar al tanque.", width/2, height/2);
    break;
  case DISPARANDO:
  case REPOSO:
    nube.mover_der((int)0.5+random(2), 0); // Movemos de izquierda a derecha la nube
    // La rana se mueve sólo cuando llega a un frame del salto en concreto
    if (rana.getFrame() == 3) rana.mover(24, 0);
    rana.draw(-1);
    // Cañón y tanque por delante de la rana
    pinta_cano();
    tanque.draw();
    // Pintamos el proyectil
    pinta_bala();

    // Miramos si las coordenadas (x,y) del proyectil están dentro de la zona del tanque
    if (tanque.dentro(x, y)) {
      // Si ha sido alcanzado cambiamos el estado a DESTRUIDO e iniciamos el contador de espera
      estado = DESTRUIDO;
      contador = TIEMPO_ESPERA;
      // Resituamos el tanque para la siguiente partida 
      tanque.pos((int)(width-178-random(300)), tanque.getY());
      tanque.reset();
      sboom.play();
    } else {
      // Si no hemos acertado al tanque lo movemos y pintamos
      tanque.mover(vtank, 0);
    }
    break;
  }
  // Pintamos el cuadro de control con esquinas redondeadas y marco de 2 pixels blanco
  fill(50, 30, 30, 200);    // Semitransparente (200)
  stroke(255, 255, 255);    // Marco blanco
  strokeWeight(2);          // Ancho marco (línea) 2 pixels
  rect(5, 5, 140, 150, 10); // Dibujar recuadro
  textSize(12);             // Tamaño del texto

  fill(255);               // Blanco del texto
  textAlign(LEFT, TOP);    // Alineación del texto
  text("Velocidad :    "+v_ini/10, 20, 12);         // Pintamos velocidad
  text("Angulo    :     "+(int)angulo+"º", 20, 32); // Pintamos el ángulo
  text("Intentos  : "+intentos, 20, 60);            // Pintamos el número de intentos

  // Pintamos los créditos de autoria
  fill(255, 255, 30); // Creditos de color AMARILLO
  text("Pedro Puertas\r\nCC-BY 2017", 20, 110);
}

/**
 Control del teclado
 */
void keyPressed() {
  // Mientras el proyectil está en el aire no se puede modificar el ángulo ni la velocidad
  if (estado == DISPARANDO) return;

  switch (keyCode) {
  case UP :    
    boton_a_mas();     
    break;
  case DOWN :  
    boton_a_menos();    
    break;
  case LEFT :  
    boton_v_menos();    
    break;
  case RIGHT : 
    boton_v_mas();    
    break;
  case 32 :    
    boton_dispara();    
    break;
  }
}

void pinta_cano() {
  image(base, 40, height-42); // Pintamos la base del cañón
  pushMatrix(); // Guardamos el estado actual del lienzo
  imageMode(CENTER); // Cambiamos el modo de posicionamiento de imagen al centro
  translate(88-angulo/4, height-60-angulo/5); // Cambiamos el punto de la posición del fusil dependiendo del ángulo marcado
  rotate(radians(-angulo)); // Se ajusta la rotación del fusil al ángulo indicado
  image(fusil, 0, 0); // Pintamos el fusil
  imageMode(CORNER);  // Volvemos a poner el mode de imagen a la esquina
  popMatrix(); // Recuperamos el estado del lienzo
}

void pinta_bala() {
  // La bala sólo se pinta si han disparado
  if (estado == DISPARANDO) {
    // Uso de la fórmula de tiro parabólico
    // x = x0 + v0*cos(ang)*t
    // y = y0 - v0*sin(angulo)*t-g*t^2/2

    // Se realiza una adaptación de la posición inicial dependiendo del ángulo que tiene el cañón
    // en la coordenada x con 130+(-angulo/2) y en la coordenada Y con (height-60-angulo)
    x= 130+(-angulo/2)+v_ini*cos(radians(angulo))*t;
    y= (height-60-angulo)-(v_ini*sin(radians(angulo))*t-gr*pow(t, 2)/2);

    // Se realiza una pequeña animación de la bala mientras está en el aire, para conseguir un efecto oscilante aprovechamos 
    // la función coseno que retorna valores -1 y 1 que multiplicados por un factor nos permite aumentar y disminuir la bala
    // ciclicamente por tanto tendremos un tamaño de la imagen entre 100% i 70% aprox
    float zoom = 0.8-(cos(t*3)/6);
    pushMatrix(); // Guardamos el estado actual
    imageMode(CENTER); // Cambiamos el modo de posición de la imagen al centro
    translate(x, y); // Cambiamos el punto origen a la posició de la bala
    rotate(radians(-angulo+t*(100/v_ini)*7)); // Ajustamos la rotación de la bala según la velocidad y el momento del vuelo
    scale(zoom); // Escalamos la imagen
    image(bala, 0, 0); // Pintamos la bala en la posició 0,0 ya que hemos trasladado el origen de coordendas
    imageMode(CORNER); // Retornamos el modo de imagen a la esquina superior izquierda
    popMatrix(); // Recuperem el estado del lienzo
    // Incrementamos la posición en el tiempo (t)
    t+=0.25;
    // Si llegamos a tierra hemos de esperar a que lancen una nueva bala, volvemos a REPOSO
    if (y > height) {
      estado = REPOSO;
      x = 0;
      intentos--; // Restamos el número de tiradas
      if (intentos <= 0) {
        // Si ya no quedan intentos se ha terminado la partida
        contador = TIEMPO_ESPERA;
        estado = ERROR;
        // Reposicionamos el tanque
        tanque.pos((int)(width-178-random(300)), tanque.getY());
        tanque.reset();
      }
    }
  }
}

// LLamadas de la libreria ControlP5 para cada botón, si estamos DISPARANDO no actuan
void boton_v_mas() {
  if (estado == DISPARANDO) return;
  if (v_ini < 80) v_ini += 5;
}  

void boton_v_menos() {
  if (estado == DISPARANDO) return;
  if (v_ini > 20) v_ini -= 5;
}

void boton_a_mas() {
  if (estado == DISPARANDO) return;
  if (angulo < 90) angulo += 5;
}

void boton_a_menos() {
  if (estado == DISPARANDO) return;
  if (angulo > 0) angulo -= 5;
}

void boton_dispara() {
  if (estado == DISPARANDO) return;
  if (intentos > 0) {
    t = 0;
    estado = DISPARANDO;
  }
}